#include "neuquant.h"
#include "..\..\common\ImageFilter.h"
#include "SettingsForm.h"

using namespace FilterSpace;
using namespace System::Windows::Forms;


bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)
	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);

	int sin = f->stridein;
	int sout = f->strideout;
	int mstride = (f->width + 7) >> 3;
	byte *din, *dout, *mask = f->mask;
	// END --> COMMON SUPPORT CODE (DO NOT MODIFY!)

	// WRITE YOUR FILTER CODE HERE -->
	if (f->channelcountin != f->channelcountout) return false;

	int processed_pixels = 0;

	int channels = f->channelcountin;

	if (mask) mask += sy * mstride;

	byte* colmap = &(f->colorlookuptable[0]);

	for (int y = 0; y < sh; ++y)
	{
		din  = f->datain + sx * channels + (sy + y) * sin;
		dout = f->dataout + y * sout;
		
		for (int x = 0; x < sw; ++x) 
		{
			int z = x + sx;
			byte b = *(din++);
			byte g = *(din++);
			byte r = *(din++);
			int  j = -1;

			if (!mask)
				j = f->neuquant->inxsearch(b, g, r);
			else if ( mask[z >> 3] & ((z & 0x07) ? (1 << (z & 0x07)) : 1) )
				j = f->neuquant->inxsearch(b, g, r);

			*(dout++) = colmap[j * 3 + 0];
			*(dout++) = colmap[j * 3 + 1];
			*(dout++) = colmap[j * 3 + 2];
		}

		if (mask) mask += mstride;

		// FOR PERFORMANCE MEASUREMENT
		processed_pixels += sw;

		if (updater != nullptr && processed_pixels > 10000) // MULTI-THREAD
		{
			if (updater->WaitOne(10, false))
			{
				*(f->processing_ptr) += processed_pixels;
				processed_pixels = 0;
				updater->Release(1);
			}
		}
		else // SINGLE-THREAD
		{
			if (f->processing_ptr) *(f->processing_ptr) += processed_pixels;
			processed_pixels = 0;
		}
		
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}
	// FILTER CODE END -->

	return true;
}







IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, stride = 0, height = 0, run_local = 1;

	if (!fs) return IntPtr(0);

	bool cancelled = false;
	xml_node^ xn = gcnew xml_node( fs );

	filtersettings* f = XmlToMyStruct( xn, true );
	//f->pixelsizeout = 24; // TEMPORARY FOR BUILT-IN IMAGEBUFFER

	if (f && !silent)
	{
		if (!f->neuquant) f->neuquant = new NeuQuant(f->neurons, f->samplingfactor, f->stridein, f->width, f->height);
		FilterTemplate::SettingsForm^ form = gcnew FilterTemplate::SettingsForm(f);
		form->RefreshPreview(true);
		form->ShowDialog();
		cancelled = form->FilterCancelled;
		delete form;
	}

	if (cancelled)
	{
		delete f->neuquant;
		delete f;
		delete xn;
		return IntPtr(0);
	}

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* fb = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(fb);
	delete [] fb;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			xt->XmlGetChildValueByName("Stride", stride);
			xt->XmlGetChildValueByName("Height", height);
			if (stride * height == 0) return IntPtr(0);
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", outdata);
		}
	}
	xn->children.Add( xt );

	xn->XmlSetChildValueByName("Neurons", f->neurons);
	xn->XmlSetChildValueByName("SamplingFactor", f->samplingfactor);
	xn->XmlSetChildValueByName("ColorMap", stringToString(base64((unsigned char*)f->colorlookuptable, 768)), false);
	
	byte* nndata;
	int   nnlength;
	if (f->neuquant->Marshall(nndata, nnlength))
		xn->XmlSetChildValueByName("NeuralNetwork", stringToString(base64(nndata, nnlength)), false);

	xn->XmlSetChildValueByName("nCore!ExecutionSize", 1); // PIXELWISE == 1 ALWAYS FOR THIS FILTER

	delete f->neuquant;
	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}


filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;
	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) return 0;

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	xn->XmlSetChildValueByName("Neurons", 256, true);
	xn->XmlGetChildValueByName("Neurons", filset->neurons);
	xn->XmlSetChildValueByName("SamplingFactor", 30, true);
	xn->XmlGetChildValueByName("SamplingFactor", filset->samplingfactor);

	byte*   data;
	int     length;
	String^ sStr;

	if (xn->XmlGetChildValueByName("ColorMap", sStr))
	{
		base64(StringTostring(sStr), data, length);
		if (length > 0) 
		{
			memcpy(filset->colorlookuptable, data, length);
			delete [] data;
		}
	}

	filset->neuquant = 0;

	if (xn->XmlGetChildValueByName("NeuralNetwork", sStr))
	{
		base64(StringTostring(sStr), data, length);
		if (length > 0) 
		{
			filset->neuquant = new NeuQuant(1, 1, 1, 1, 1);
			filset->neuquant->DeMarshall(data, length);
			delete [] data;
		}

	}

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) ) ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) ) ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;

	return filset;
}



bool ImageFilter::CanLocalize(){ return true; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            3,        24,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	SetChannelsBitAllocation(xn, 1, 8);
	SetChannelsBitAllocation(xn, 2, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName,  width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,            3,        24,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	SetChannelsBitAllocation(xn, 1, 8);
	SetChannelsBitAllocation(xn, 2, 8);

	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}


// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		//MessageBox::Show(((int)filset->processing_ptr).ToString() + " " + ((int)filset->processing_size).ToString());

		*(filset->processing_ptr)  = 0;
		
		*(filset->processing_size) = filset->width * filset->height;

		if (filset->neuquant == 0)
		{
			MessageBox::Show("NeuralNetwork not found!!!");
			delete filset;
			return false;
		}

		int cpucores = System::Environment::ProcessorCount;

		if (!filset->nCoreEnabled || cpucores == 1)
		{
			b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
		}
		else
		{
			filter_core_func f = (::FilterCore);

			if (FILTER_TYPE == IMAGEFILTER_PIXELWISE)
			{
				filset->nCoreExecutionSize = 1; // SAFETY CONDITIONING
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
			else if (FILTER_TYPE == IMAGEFILTER_BOX)
			{
				b = ImageFilterBase::ExecuteFilterBox(filset, f);
			}
			else if (FILTER_TYPE == IMAGEFILTER_BLOCK)
			{
				b = ImageFilterBase::ExecuteFilterBlock(filset, f);
			}
			else // if (FILTER_TYPE == IMAGEFILTER_RECURSIVE) AND OTHERS
			{
				// NO AUTOMATIC MULTITHREAD SUPPORT
				b = ::FilterCore(filset, 0, 0, 0, 0, nullptr); // FASTCALL WITH 0,0,0,0 PARAMETERS
			}
		}
		delete filset->neuquant;
		delete filset;
		*(filset->processing_ptr)  = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}
