#pragma once

using namespace System;
using namespace System::ComponentModel;
using namespace System::Collections;
using namespace System::Windows::Forms;
using namespace System::Data;
using namespace System::Drawing;
using namespace System::Drawing::Imaging;


namespace FilterTemplate {

	public ref class SettingsForm : public System::Windows::Forms::Form
	{

	public:
		filtersettings* FilterSetting;
		bool FilterCancelled, MouseButtonDown;
		Bitmap^ PreviewBitmap;
		Bitmap^ PaletteBitmap;
		Rectangle SourceRect, DestinationRect;
		int SWidth, SHeight, LastX, LastY;
		int fromIndex;

	private: System::Windows::Forms::Label^			label1;
	private: System::Windows::Forms::Label^			label2;
	private: System::Windows::Forms::HScrollBar^	ScrollBarNeurons;
	private: System::Windows::Forms::HScrollBar^	ScrollBarSamplingFactor;
	private: System::Windows::Forms::PictureBox^	Spacer;
	private: System::Windows::Forms::PictureBox^	PictureBoxColorMap;
	private: System::Windows::Forms::PictureBox^	pictureBox1;

	private: System::Windows::Forms::Label^  label3;
	private: System::Windows::Forms::Button^		ButtonAbout;



	public:
		SettingsForm(filtersettings* fs)
		{
			fromIndex = -1;
			FilterSetting = fs;
			FilterSetting->neuquant->initnet(FilterSetting->datain, FilterSetting->width * FilterSetting->height * 3);
			FilterSetting->neuquant->learn();
			FilterSetting->neuquant->unbiasnet();
			FilterSetting->neuquant->putcolourmaptomem((byte*)(FilterSetting->colorlookuptable));
			FilterSetting->neuquant->inxbuild();

			InitializeComponent();

			ScrollBarNeurons->Minimum = 16;
			ScrollBarNeurons->Maximum = 255;
			ScrollBarSamplingFactor->Minimum = 1;
			ScrollBarSamplingFactor->Maximum = 30;
			if (FilterSetting->neurons <= ScrollBarNeurons->Maximum && ScrollBarNeurons->Minimum >= FilterSetting->neurons) ScrollBarNeurons->Value = FilterSetting->neurons;
			if (FilterSetting->samplingfactor <= ScrollBarSamplingFactor->Maximum && ScrollBarSamplingFactor->Minimum >= FilterSetting->samplingfactor) ScrollBarSamplingFactor->Value = FilterSetting->samplingfactor;

			SWidth = fs->width;
			SHeight = fs->height;

			FilterCancelled = true;
			PreviewBitmap = gcnew Bitmap(Preview->ClientSize.Width, Preview->ClientSize.Height, PixelFormat::Format24bppRgb);
			PaletteBitmap = gcnew Bitmap(PictureBoxColorMap->ClientSize.Width, PictureBoxColorMap->ClientSize.Height, PixelFormat::Format24bppRgb);

			DestinationRect.X = (SWidth < Preview->ClientSize.Width) ? ((Preview->ClientSize.Width - SWidth) / 2) : 0;
			DestinationRect.Y = (SHeight < Preview->ClientSize.Height) ? ((Preview->ClientSize.Height - SHeight) / 2) : 0;
			DestinationRect.Width = (DestinationRect.X) ? SWidth : Preview->ClientSize.Width;
			DestinationRect.Height = (DestinationRect.Y) ? SHeight : Preview->ClientSize.Height;

			SourceRect.X = 0;
			SourceRect.Y = 0;
			SourceRect.Width = DestinationRect.Width;   // SHOULD BE EQUAL !!!
			SourceRect.Height = DestinationRect.Height; // SHOULD BE EQUAL !!!

			MouseButtonDown = false;

			ScrollBar_Scroll(nullptr, nullptr);
		}

	protected:

		~SettingsForm()
		{
			if (components)
			{
				delete components;
			}
		}


	private: System::Windows::Forms::Button^  ButtonApply;
		System::Windows::Forms::Button^  ButtonCancel;
		System::Windows::Forms::PictureBox^  Preview;
		System::ComponentModel::IContainer^  components;
	

#pragma region Windows Form Designer generated code

		void InitializeComponent(void)
		{
			this->Preview = (gcnew System::Windows::Forms::PictureBox());
			this->ButtonApply = (gcnew System::Windows::Forms::Button());
			this->ButtonCancel = (gcnew System::Windows::Forms::Button());
			this->label1 = (gcnew System::Windows::Forms::Label());
			this->ButtonAbout = (gcnew System::Windows::Forms::Button());
			this->label2 = (gcnew System::Windows::Forms::Label());
			this->ScrollBarNeurons = (gcnew System::Windows::Forms::HScrollBar());
			this->ScrollBarSamplingFactor = (gcnew System::Windows::Forms::HScrollBar());
			this->Spacer = (gcnew System::Windows::Forms::PictureBox());
			this->PictureBoxColorMap = (gcnew System::Windows::Forms::PictureBox());
			this->pictureBox1 = (gcnew System::Windows::Forms::PictureBox());
			this->label3 = (gcnew System::Windows::Forms::Label());
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Preview))->BeginInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Spacer))->BeginInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->PictureBoxColorMap))->BeginInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->pictureBox1))->BeginInit();
			this->SuspendLayout();
			// 
			// Preview
			// 
			this->Preview->BorderStyle = System::Windows::Forms::BorderStyle::Fixed3D;
			this->Preview->Cursor = System::Windows::Forms::Cursors::NoMove2D;
			this->Preview->Location = System::Drawing::Point(12, 12);
			this->Preview->Name = L"Preview";
			this->Preview->Size = System::Drawing::Size(154, 154);
			this->Preview->TabIndex = 0;
			this->Preview->TabStop = false;
			this->Preview->MouseMove += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::Preview_MouseMove);
			this->Preview->MouseDown += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::Preview_MouseDown);
			this->Preview->Paint += gcnew System::Windows::Forms::PaintEventHandler(this, &SettingsForm::Preview_OnPaint);
			this->Preview->MouseUp += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::Preview_MouseUp);
			// 
			// ButtonApply
			// 
			this->ButtonApply->Location = System::Drawing::Point(183, 143);
			this->ButtonApply->Name = L"ButtonApply";
			this->ButtonApply->Size = System::Drawing::Size(53, 23);
			this->ButtonApply->TabIndex = 2;
			this->ButtonApply->Text = L"Apply";
			this->ButtonApply->UseVisualStyleBackColor = true;
			this->ButtonApply->Click += gcnew System::EventHandler(this, &SettingsForm::ButtonApply_Click);
			// 
			// ButtonCancel
			// 
			this->ButtonCancel->Location = System::Drawing::Point(242, 143);
			this->ButtonCancel->Name = L"ButtonCancel";
			this->ButtonCancel->Size = System::Drawing::Size(51, 23);
			this->ButtonCancel->TabIndex = 3;
			this->ButtonCancel->Text = L"Cancel";
			this->ButtonCancel->UseVisualStyleBackColor = true;
			this->ButtonCancel->Click += gcnew System::EventHandler(this, &SettingsForm::ButtonCancel_Click);
			// 
			// label1
			// 
			this->label1->AutoSize = true;
			this->label1->ForeColor = System::Drawing::Color::White;
			this->label1->Location = System::Drawing::Point(180, 12);
			this->label1->Name = L"label1";
			this->label1->Size = System::Drawing::Size(172, 13);
			this->label1->TabIndex = 10;
			this->label1->Text = L"Number of colors (Neurons 16-255)";
			// 
			// ButtonAbout
			// 
			this->ButtonAbout->Location = System::Drawing::Point(299, 143);
			this->ButtonAbout->Name = L"ButtonAbout";
			this->ButtonAbout->Size = System::Drawing::Size(53, 23);
			this->ButtonAbout->TabIndex = 26;
			this->ButtonAbout->Text = L"About";
			this->ButtonAbout->UseVisualStyleBackColor = true;
			this->ButtonAbout->Click += gcnew System::EventHandler(this, &SettingsForm::ButtonAbout_Click);
			// 
			// label2
			// 
			this->label2->AutoSize = true;
			this->label2->ForeColor = System::Drawing::Color::White;
			this->label2->Location = System::Drawing::Point(180, 77);
			this->label2->Name = L"label2";
			this->label2->Size = System::Drawing::Size(113, 13);
			this->label2->TabIndex = 27;
			this->label2->Text = L"Sampling Factor (1-30)";
			// 
			// ScrollBarNeurons
			// 
			this->ScrollBarNeurons->LargeChange = 1;
			this->ScrollBarNeurons->Location = System::Drawing::Point(183, 35);
			this->ScrollBarNeurons->Maximum = 255;
			this->ScrollBarNeurons->Minimum = 16;
			this->ScrollBarNeurons->Name = L"ScrollBarNeurons";
			this->ScrollBarNeurons->Size = System::Drawing::Size(199, 17);
			this->ScrollBarNeurons->TabIndex = 28;
			this->ScrollBarNeurons->Value = 16;
			this->ScrollBarNeurons->Scroll += gcnew System::Windows::Forms::ScrollEventHandler(this, &SettingsForm::ScrollBar_Scroll);
			// 
			// ScrollBarSamplingFactor
			// 
			this->ScrollBarSamplingFactor->LargeChange = 1;
			this->ScrollBarSamplingFactor->Location = System::Drawing::Point(183, 101);
			this->ScrollBarSamplingFactor->Maximum = 30;
			this->ScrollBarSamplingFactor->Minimum = 1;
			this->ScrollBarSamplingFactor->Name = L"ScrollBarSamplingFactor";
			this->ScrollBarSamplingFactor->Size = System::Drawing::Size(199, 17);
			this->ScrollBarSamplingFactor->TabIndex = 29;
			this->ScrollBarSamplingFactor->Value = 30;
			this->ScrollBarSamplingFactor->Scroll += gcnew System::Windows::Forms::ScrollEventHandler(this, &SettingsForm::ScrollBar_Scroll);
			// 
			// Spacer
			// 
			this->Spacer->BackColor = System::Drawing::Color::Orange;
			this->Spacer->Location = System::Drawing::Point(391, 12);
			this->Spacer->Name = L"Spacer";
			this->Spacer->Size = System::Drawing::Size(3, 154);
			this->Spacer->TabIndex = 30;
			this->Spacer->TabStop = false;
			// 
			// PictureBoxColorMap
			// 
			this->PictureBoxColorMap->BorderStyle = System::Windows::Forms::BorderStyle::FixedSingle;
			this->PictureBoxColorMap->Location = System::Drawing::Point(400, 35);
			this->PictureBoxColorMap->Name = L"PictureBoxColorMap";
			this->PictureBoxColorMap->Size = System::Drawing::Size(330, 131);
			this->PictureBoxColorMap->TabIndex = 32;
			this->PictureBoxColorMap->TabStop = false;
			this->PictureBoxColorMap->MouseDown += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::PictureBoxColorMap_MouseDown);
			this->PictureBoxColorMap->Paint += gcnew System::Windows::Forms::PaintEventHandler(this, &SettingsForm::PictureBoxColorMap_Paint);
			// 
			// pictureBox1
			// 
			this->pictureBox1->BackColor = System::Drawing::Color::Orange;
			this->pictureBox1->Location = System::Drawing::Point(172, 12);
			this->pictureBox1->Name = L"pictureBox1";
			this->pictureBox1->Size = System::Drawing::Size(3, 154);
			this->pictureBox1->TabIndex = 33;
			this->pictureBox1->TabStop = false;
			// 
			// label3
			// 
			this->label3->AutoSize = true;
			this->label3->ForeColor = System::Drawing::Color::White;
			this->label3->Location = System::Drawing::Point(397, 12);
			this->label3->Name = L"label3";
			this->label3->Size = System::Drawing::Size(243, 13);
			this->label3->TabIndex = 35;
			this->label3->Text = L"Left click to choose color, Right click to overwrite.";
			// 
			// SettingsForm
			// 
			this->AutoScaleDimensions = System::Drawing::SizeF(6, 13);
			this->AutoScaleMode = System::Windows::Forms::AutoScaleMode::Font;
			this->BackColor = System::Drawing::Color::SteelBlue;
			this->ClientSize = System::Drawing::Size(742, 178);
			this->Controls->Add(this->label3);
			this->Controls->Add(this->pictureBox1);
			this->Controls->Add(this->PictureBoxColorMap);
			this->Controls->Add(this->Spacer);
			this->Controls->Add(this->ScrollBarSamplingFactor);
			this->Controls->Add(this->ScrollBarNeurons);
			this->Controls->Add(this->label2);
			this->Controls->Add(this->ButtonAbout);
			this->Controls->Add(this->label1);
			this->Controls->Add(this->ButtonCancel);
			this->Controls->Add(this->ButtonApply);
			this->Controls->Add(this->Preview);
			this->FormBorderStyle = System::Windows::Forms::FormBorderStyle::FixedSingle;
			this->MaximizeBox = false;
			this->MinimizeBox = false;
			this->Name = L"SettingsForm";
			this->ShowIcon = false;
			this->ShowInTaskbar = false;
			this->StartPosition = System::Windows::Forms::FormStartPosition::CenterScreen;
			this->Text = L"NeuQuantPlus + Color-remap Adjust";
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Preview))->EndInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Spacer))->EndInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->PictureBoxColorMap))->EndInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->pictureBox1))->EndInit();
			this->ResumeLayout(false);
			this->PerformLayout();

		}
#pragma endregion

	public:
		void RefreshPreview(bool applyfilter)
		{
			BitmapData^ bd = PreviewBitmap->LockBits( Rectangle(0, 0, PreviewBitmap->Width, PreviewBitmap->Height), ImageLockMode::ReadWrite, PixelFormat::Format24bppRgb);

			FilterSetting->strideout = bd->Stride;
			FilterSetting->dataout = (Byte*)(void*)bd->Scan0;

			if (applyfilter)
				FilterSpace::ImageFilter::FilterCore(FilterSetting, SourceRect.X, SourceRect.Y, SourceRect.Width, SourceRect.Height);
			else
				FilterSpace::ImageFilter::ClipImage( FilterSetting, SourceRect.X, SourceRect.Y, SourceRect.Width, SourceRect.Height);

			PreviewBitmap->UnlockBits(bd);
			this->Refresh();
		}

	private: 
		System::Void ButtonApply_Click(System::Object^  sender, System::EventArgs^  e) 
		{
			FilterCancelled = false;
			this->Close();
		}

		System::Void ButtonCancel_Click(System::Object^  sender, System::EventArgs^  e) 
		{
			this->Close();
		}

		System::Void Preview_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			if (e->Button == System::Windows::Forms::MouseButtons::Left) 
			{
				MouseButtonDown = true;
				LastX = e->X;
				LastY = e->Y;
				RefreshPreview(false);
			}
		}

		System::Void Preview_MouseMove(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			if (!MouseButtonDown) return;
			if (!DestinationRect.X)
			{
				SourceRect.X -= e->X - LastX;
				if (SourceRect.X < 0) SourceRect.X = 0;
				if (SourceRect.X >= SWidth - DestinationRect.Width) SourceRect.X = SWidth - DestinationRect.Width - 1;
			}
			if (!DestinationRect.Y)
			{
				SourceRect.Y -= e->Y - LastY;
				if (SourceRect.Y < 0) SourceRect.Y = 0;
				if (SourceRect.Y >= SHeight - DestinationRect.Height) SourceRect.Y = SHeight - DestinationRect.Height - 1;
			}
			RefreshPreview(false);
			LastX = e->X;
			LastY = e->Y;
		}

		System::Void Preview_MouseUp(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			if (e->Button == System::Windows::Forms::MouseButtons::Left) MouseButtonDown = false;
			if (!MouseButtonDown) RefreshPreview(true);
		}

		System::Void Preview_OnPaint(System::Object^  sender, System::Windows::Forms::PaintEventArgs^  e) 
		{
			Graphics^ g = e->Graphics;
			Rectangle r = Rectangle(0, 0, DestinationRect.Width, DestinationRect.Height);
			g->DrawImage(PreviewBitmap, DestinationRect, r, GraphicsUnit::Pixel);
		}

		void PictureBoxColorMap_Paint(System::Object^  sender, System::Windows::Forms::PaintEventArgs^  e) 
		{
			Graphics^ g = e->Graphics;
			Rectangle r = Rectangle(0, 0, PaletteBitmap->Width, PaletteBitmap->Height);
			g->DrawImage(PaletteBitmap, r, r, GraphicsUnit::Pixel);
		}

		void ButtonAbout_Click(System::Object^  sender, System::EventArgs^  e) 
		{
			String^ msg = "NeuQuant Neural-Net Quantization Algorithm Interface\n\n";
			msg        += "NEUQUANT Neural-Net quantization algorithm by Anthony Dekker, 1994.\n";
			msg        += "See \"Kohonen neural networks for optimal colour quantization\"\n";
			msg        += "in \"Network: Computation in Neural Systems\" Vol. 5 (1994) pp 351-367.\n";
			msg        += "for a discussion of the algorithm.\n\n";
			msg        += "See also  http://members.ozemail.com.au/~dekker/NEUQUANT.HTML\n\n";
			msg        += "This project is ported to Managed C++ and extended with color table\n";
			msg        += "remapping by Rudolf Szendrei, 2009.\n";

			String^ cap = "About NeuQuant Plus";
			MessageBox::Show(msg, cap, System::Windows::Forms::MessageBoxButtons::OK, System::Windows::Forms::MessageBoxIcon::Information);
		}

		void ScrollBar_Scroll(System::Object^  sender, System::Windows::Forms::ScrollEventArgs^  e)
		{
			FilterSetting->neurons = ScrollBarNeurons->Value;
			FilterSetting->samplingfactor = ScrollBarSamplingFactor->Value;
			label1->Text = "Number of colors (Neurons 16-255): " + ScrollBarNeurons->Value.ToString();
			label2->Text = "Sampling Factor (1-30): " + ScrollBarSamplingFactor->Value.ToString();

			FilterSetting->neuquant->SetNewParameters(FilterSetting->neurons, FilterSetting->samplingfactor);
			FilterSetting->neuquant->initnet(FilterSetting->datain, FilterSetting->width * FilterSetting->height * 3);
			FilterSetting->neuquant->learn();
			FilterSetting->neuquant->unbiasnet();
			FilterSetting->neuquant->putcolourmaptomem(&(FilterSetting->colorlookuptable[0]));
			FilterSetting->neuquant->inxbuild();

			RefreshPalette();
		}

		void RefreshPalette()
		{
			Graphics^ gr    = Graphics::FromImage(PaletteBitmap);

			int   height      = PaletteBitmap->Height;
			int   width       = PaletteBitmap->Width;
			int   blockwidth  = width / 32;
			int   blockheight = height / 8;
			int   index       = 0;

			Pen^   pen   = gcnew Pen(Color::White);
			Brush^ brush = gcnew SolidBrush(Color::FromArgb(192, 192, 192));
			gr->FillRectangle(brush, Rectangle(0, 0, width, height));
			delete brush;

			for (int by = 0; by < 8; by++)
			{
				for (int bx = 0; bx < 32; bx++)
				{
					byte b = FilterSetting->colorlookuptable[index * 3 + 0];
					byte g = FilterSetting->colorlookuptable[index * 3 + 1];
					byte r = FilterSetting->colorlookuptable[index * 3 + 2];
		
					brush = gcnew SolidBrush(Color::FromArgb(r, g, b));
					gr->FillRectangle(brush, Rectangle(bx * blockwidth + 1, by * blockheight + 1, blockwidth - 2, blockheight - 2));
					gr->DrawRectangle(pen, Rectangle(bx * blockwidth + 1, by * blockheight + 1, blockwidth - 2, blockheight - 2));
					delete brush;

					index++;
					if (index == FilterSetting->neurons) break;
				}
				if (index == FilterSetting->neurons) break;
			}

			PictureBoxColorMap->Refresh();			
		}

		void PictureBoxColorMap_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			int   blockwidth  = PaletteBitmap->Width / 32;
			int   blockheight = PaletteBitmap->Height / 8;
			int   x     = e->X / blockwidth;
			int   y     = e->Y / blockheight;
			int   index = y * 32 + x;

			if (e->Button == System::Windows::Forms::MouseButtons::Left)
			{
				fromIndex = index;
			}
			else
			{
				if (fromIndex != -1)
				{
					FilterSetting->colorlookuptable[index * 3 + 0] = FilterSetting->colorlookuptable[fromIndex * 3 + 0];
					FilterSetting->colorlookuptable[index * 3 + 1] = FilterSetting->colorlookuptable[fromIndex * 3 + 1];
					FilterSetting->colorlookuptable[index * 3 + 2] = FilterSetting->colorlookuptable[fromIndex * 3 + 2];
				}
			}
			RefreshPalette();
			Refresh();
		}
};
}
